/*
Programm: Aufg.Klausur.2000-03-09.03
Zweck:    counts the lines of output from ls -l
Aufgabe:  This is task 3 in the exam from 2000-03-09 in Betriebssysteme 1 with
          Prof. Dr. Michael Jaeger at the University of Applied Sciences Giessen-Friedberg:
Autor:    Mattias Ansorg <matthias@ansorgs.de>
Lizenz:   public domain
Datum:    2002-11-16
*/

#include <unistd.h> //fork(), read(), pipe()
#include <signal.h> //sigaction, SIGCHLD, sigemptyset()
#include <sys/wait.h> //wait()
#include <stdio.h>  //perror()
#include <string.h> //strchr()
#include <iostream> //cin, cout
#include <string>

using namespace std;

//for some reason we dont need extern "C" { }; enclosing for this SIGCHLD handler
//We need this handler to acknowledge that a child is terminated otherwise it 
//will remain as a zombie process. This is not necessary for pipe programming,
//as a zombie process has no open file descriptors any more.

void chldHandler(int signum) { //SIGCHLD handler 
  //we know that a child is now a zombie. By wait() 
  //we make it terminate finally, without knowing the childs pid 
  wait(0); 
}

int main() {
  int forkRes, pipeRes, dup2Res, lineCount=0;
  int thePipe[2];
  int &readPipe  = thePipe[0]; // caution: these are reference variables
  int &writePipe = thePipe[1]; 
  ssize_t bytesRead;
  char input[241]; //portion of input
  char *inputPtr;  //point to a position in input, i.e. specify a substring
  struct sigaction oldAction, newAction;

  newAction.sa_handler = &chldHandler; //signal handler function 
  newAction.sa_flags = 0; 
  sigemptyset(&newAction.sa_mask); //ignore no signals 
  if (sigaction(SIGCHLD, &newAction, &oldAction) ==-1) 
    perror("main(): error occured on sigaction()\n");
  
  pipeRes = pipe(thePipe); //create Pipe; hereafter, readPipe is read descriptor, writePipe is write descriptor
  if (pipeRes == -1) { 
    perror("main() : error occured on pipe()\n");
    exit(1);
  }

  switch (forkRes=fork()) {
    case -1: perror("main() : error occured on fork()\n");
      break;
    case  0: //means we are in child process
      dup2Res = dup2(writePipe,1); //io-redirection of stdout to pipe
      if (dup2Res != -1) {
	close(writePipe);
	sleep(1);
        execlp("ls", "ls", "-l", 0);
        perror("main() : error occured on execlp()"); //execlp returned? means, an error occured!
	exit(1);
      }
      else {
	perror("main(): error occured on dup2() in child process\n");
	//cout << "tried to use file descriptor writePipe == " << writePipe << endl;
      }
      break;
    default: //means we are in parent process
      close(writePipe); //this is crucial; a read() terminates only if there is no open write 
                        //descriptor of the pipe at all; so, if we don't close it here before read(),
                        //we're going to listen to ourselves forever while read()ing sleeps.
      dup2Res = dup2(readPipe,0); //io-redirection of stdin from pipe
      if (dup2Res != -1) {
        do {
  	  bytesRead = read(readPipe,input,240);
	  input[bytesRead] = '\0'; //terminate this string after bytesRead bytes
	  //printf("\n%s\n", input); //for testing; must be after string termination!
          inputPtr = input; 
	  //cout << endl << "debug: bytesRead = " << bytesRead << endl; //debug code
          do {
	    inputPtr = strchr(++inputPtr, '\n'); //strchr returns 0 if not found before end-of-string
	    if (inputPtr!=0)
	      ++lineCount;
          } while (inputPtr!=0);
        } while (bytesRead!=0); //0 means EOF reached, resp. for pipes: writer died
	lineCount -= 1; //ignore very last line break
        cout << "number of lines in ls -l output: " << lineCount << endl;
	close(readPipe);
      }
      else {
	perror("main() : error occured on dup2() in main process\n");
      }
      break;
  }
}
